package ru.dbotthepony.kstarbound.server

import com.google.gson.JsonObject
import io.netty.channel.ChannelHandlerContext
import it.unimi.dsi.fastutil.bytes.ByteArrayList
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectLinkedOpenHashSet
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.io.ByteKey
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.client.network.packets.ForgetChunkPacket
import ru.dbotthepony.kstarbound.client.network.packets.ChunkCellsPacket
import ru.dbotthepony.kstarbound.client.network.packets.ForgetEntityPacket
import ru.dbotthepony.kstarbound.client.network.packets.SpawnWorldObjectPacket
import ru.dbotthepony.kstarbound.network.Connection
import ru.dbotthepony.kstarbound.network.ConnectionSide
import ru.dbotthepony.kstarbound.network.ConnectionType
import ru.dbotthepony.kstarbound.network.IServerPacket
import ru.dbotthepony.kstarbound.network.packets.ClientContextUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.LegacyTileArrayUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.ServerDisconnectPacket
import ru.dbotthepony.kstarbound.server.world.WorldStorage
import ru.dbotthepony.kstarbound.server.world.LegacyWorldStorage
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.IChunkListener
import ru.dbotthepony.kstarbound.world.api.ImmutableCell
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.WorldObject
import kotlin.properties.Delegates

// serverside part of connection
class ServerConnection(val server: StarboundServer, type: ConnectionType) : Connection(ConnectionSide.SERVER, type) {
	var world: ServerWorld? = null
	lateinit var shipWorld: ServerWorld
		private set

	var skyVersion = 0L

	init {
		connectionID = server.channels.nextConnectionID()

		rpc.add("team.fetchTeamStatus") {
			JsonObject()
		}
	}

	override fun toString(): String {
		val channel = if (hasChannel) channel.remoteAddress().toString() else "<no channel>"
		val ship = if (::shipWorld.isInitialized) shipWorld.toString() else "<no shipworld>"
		return "ServerConnection[ID=$connectionID channel=$channel / $ship]"
	}

	var trackedPosition: Vector2d = Vector2d.ZERO
		set(value) {
			if (field != value) {
				field = value
				needsToRecomputeTrackedChunks = true
			}
		}

	var trackedPositionChunk: ChunkPos = ChunkPos.ZERO
		private set

	var trackedChunksWidth = 4
		set(value) {
			if (field != value) {
				field = value
				needsToRecomputeTrackedChunks = true
			}
		}

	var trackedChunksHeight = 4
		set(value) {
			if (field != value) {
				field = value
				needsToRecomputeTrackedChunks = true
			}
		}

	private val shipChunks = Object2ObjectOpenHashMap<ByteKey, KOptional<ByteArray>>()
	private val modifiedShipChunks = ObjectOpenHashSet<ByteKey>()
	var shipChunkSource by Delegates.notNull<WorldStorage>()
		private set

	override fun setupLegacy() {
		super.setupLegacy()
		shipChunkSource = LegacyWorldStorage.memory(shipChunks)
	}

	override fun setupNative() {
		super.setupNative()
		shipChunkSource = WorldStorage.EMPTY
	}

	fun receiveShipChunks(chunks: Map<ByteKey, KOptional<ByteArray>>) {
		check(shipChunks.isEmpty()) { "Already has ship chunks" }
		shipChunks.putAll(chunks)
	}

	private val tickets = Object2ObjectOpenHashMap<ChunkPos, ServerWorld.ITicket>()
	private val pendingSend = ObjectLinkedOpenHashSet<ChunkPos>()

	private var needsToRecomputeTrackedChunks = true

	private inner class ChunkListener(val pos: ChunkPos) : IChunkListener {
		override fun onEntityAdded(entity: AbstractEntity) {
			if (entity is WorldObject)
				send(SpawnWorldObjectPacket(entity.uuid, entity.serialize()))
		}

		override fun onEntityRemoved(entity: AbstractEntity) {
			send(ForgetEntityPacket(entity.uuid))
		}

		override fun onCellChanges(x: Int, y: Int, cell: ImmutableCell) {
			pendingSend.add(pos)
		}
	}

	override fun flush() {
		val entries = rpc.write()

		if (entries != null || modifiedShipChunks.isNotEmpty()) {
			channel.write(ClientContextUpdatePacket(
				entries ?: listOf(),
				KOptional(modifiedShipChunks.associateWith { shipChunks[it]!! }),
				KOptional(ByteArrayList())))

			modifiedShipChunks.clear()
		}

		super.flush()
	}

	fun onLeaveWorld() {
		tickets.values.forEach { it.cancel() }
		tickets.clear()
		pendingSend.clear()
	}

	override fun onChannelClosed() {
		super.onChannelClosed()
		server.channels.freeConnectionID(connectionID)
		server.channels.connections.remove(this)
		server.freeNickname(nickname)

		announceDisconnect("Connection to remote host is lost.")

		if (::shipWorld.isInitialized) {
			shipWorld.close()
		}
	}

	private var announcedDisconnect = false

	private fun announceDisconnect(reason: String) {
		if (!announcedDisconnect && nickname.isNotBlank()) {
			if (reason.isBlank()) {
				server.chat.systemMessage("Player '$nickname' disconnected")
			} else {
				server.chat.systemMessage("Player '$nickname' disconnected ($reason)")
			}

			announcedDisconnect = true
		}
	}

	override fun disconnect(reason: String) {
		announceDisconnect(reason)

		if (channel.isOpen) {
			// send pending updates
			flush()
		}

		tickets.values.forEach { it.cancel() }
		tickets.clear()
		pendingSend.clear()

		if (::shipWorld.isInitialized) {
			shipWorld.close()
		}

		if (channel.isOpen) {
			// say goodbye
			channel.write(ServerDisconnectPacket(reason))
			channel.flush()
			channel.close()
		}
	}

	private fun recomputeTrackedChunks() {
		val world = world ?: return
		val trackedPositionChunk = world.geometry.chunkFromCell(trackedPosition)
		needsToRecomputeTrackedChunks = false
		// if (trackedPositionChunk == this.trackedPositionChunk) return

		val tracked = ObjectOpenHashSet<ChunkPos>()

		for (x in -trackedChunksWidth .. trackedChunksWidth) {
			for (y in -trackedChunksHeight .. trackedChunksHeight) {
				tracked.add(world.geometry.wrap(trackedPositionChunk + ChunkPos(x, y)))
			}
		}

		val itr = tickets.entries.iterator()

		for ((pos, ticket) in itr) {
			if (pos !in tracked) {
				send(ForgetChunkPacket(pos))
				pendingSend.remove(pos)
				ticket.cancel()
				itr.remove()
			}
		}

		for (pos in tracked) {
			if (pos !in tickets) {
				val ticket = world.permanentChunkTicket(pos)
				tickets[pos] = ticket
				ticket.listener = ChunkListener(pos)
				pendingSend.add(pos)
			}
		}
	}

	fun tick() {
		val world = world

		if (world == null) {
			onLeaveWorld()
			return
		}

		if (needsToRecomputeTrackedChunks) {
			recomputeTrackedChunks()
		}

		val itr = pendingSend.iterator()

		for (pos in itr) {
			val chunk = world.chunkMap[pos] ?: continue

			if (isLegacy) {
				send(LegacyTileArrayUpdatePacket(chunk))
			} else {
				send(ChunkCellsPacket(chunk))
			}

			itr.remove()
		}
	}

	override fun channelRead(ctx: ChannelHandlerContext, msg: Any) {
		if (msg is IServerPacket) {
			try {
				msg.play(this)
			} catch (err: Throwable) {
				LOGGER.error("Failed to read serverbound packet $msg", err)
				disconnect(err.toString())
			}
		} else {
			LOGGER.error("Unknown serverbound packet type $msg")
			disconnect("Unknown serverbound packet type $msg")
		}
	}

	override fun inGame() {
		server.chat.systemMessage("Player '$nickname' connected")

		if (!isLegacy) {
			server.playerInGame(this)
		} else {
			LOGGER.info("Initializing ship world for $this")

			ServerWorld.load(server, shipChunkSource).thenAccept {
				shipWorld = it
				shipWorld.thread.start()
				shipWorld.acceptPlayer(this)
			}.exceptionally {
				LOGGER.error("Error while initializing shipworld for $this", it)
				disconnect("Error while initializing shipworld for player: $it")
				null
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
